
import re
import time
import traceback
from RestrictedPython import (compile_restricted, safe_globals, utility_builtins,)
from RestrictedPython.PrintCollector import PrintCollector
#
from .ea_node_transformer import EARestrictingNodeTransformer

import logging
log = logging.getLogger(__name__)




#---------------------------------------------------------------------------------------------------------------------
# EABase Classes
#---------------------------------------------------------------------------------------------------------------------
class EABase():
    """"""
    def __init__(self, params):
        self.language = 'en'
        self.symbol = None
        self.byte_code = None
        self.script = None
        self.script_path = None
        self.script_settings = None
        self.script_libs = []
        self.last_exec_time = 0
        self.safe_globals = safe_globals.copy()
        self.global_values = params.get('global_values', {})
        self.loc = {}
        self.df_columns = ('t', 's', 'o', 'h', 'l', 'c', 'v', 'a', 'b')
        self.ea_log_callback = params.get('ea_log_callback', None)
        #

    def get_print_factory(self, _getattr_=None):
        """"""
        if self.print_collection is None:
            self.print_collection = PrintCollector(_getattr_)
        return self.print_collection

    # parseScript(scriptText){
    #   //testerForm.script
    #   let ret = {}
    #   if(!scriptText)
    #     return ret
    #   let arrayOfLines = scriptText.match(/[^\r\n]+/g)
    #   for (var idx in arrayOfLines){
    #     let line = arrayOfLines[idx]
    #     if(!line)
    #       break
    #     if(!line.startsWith('###'))
    #       break
    #     let ary = line.split('=')
    #     if(ary.length === 2){
    #       const l = ['name', 'version', 'label', 'script_settings']
    #       let beginText = ary[0]
    #       //###[name]=DT-MA
    #       //###[version]=1.8
    #       //###[label]=timeout: 5min, near:4h, far:24h, open: ma-breakout, close: ma+c_stoploss, tp:profit+pips
    #       //###[script_settings]
    #       for(var li in l){
    #         if(beginText === `###[${l[li]}]`){
    #           ret[l[li]] = ary[1]
    #           break
    #         }
    #       }
    #     }
    #   }
    #   return ret
    # }
    def lib_path(self, name, version):
        return f"{name}_V{version}.py"

    def load_libs(self, lib_list):
        # ["PriceAction==0.1.0"]
        ret = []
        for lib in lib_list:
            lib_name, lib_version = lib.split("==")
            lib_path = self.lib_path(lib_name, lib_version)
            code = open(lib_path).read()
            lib_metadata = self.parse_script(code)
            lib_data = dict(name=lib_name, version=lib_version, path=lib_path, code=code, metadata=lib_metadata)
            ret.append(lib_data)
        return ret

    @staticmethod
    def parse_script(script_text):
        ret = {}
        try:
            if not script_text or not isinstance(script_text, str):
                return ret
            keywords = ['author', 'copyright', 'name', 'version', 'label', 'script_settings', 'lib', 'library']
            array_of_lines = script_text.splitlines()
            for line in array_of_lines:
                if not line:
                    break
                tags = re.findall(r"^###\[\w+]=", line)
                if len(tags) == 0:
                    break
                for kw in keywords:
                    lh = f"###[{kw}]="
                    if tags[0] == lh:
                        ret[kw] = line[len(tags[0]):].strip()
                        break
        except:
            traceback.print_exc()
        return ret

    def get_param(self, name, default=None):
        try:
            if self.script_settings and isinstance(self.script_settings, dict):
                params = self.script_settings.get("params", {})
                if name in params.keys():
                    params_value = params[name]
                    if params_value is None or not isinstance(params_value, dict):
                        return params_value
                    if "config" in params_value.keys():
                        return params_value.get("value", default)
                    else:
                        return params_value
        except:
            traceback.print_exc()
        return default
    #
    def get_settings(self, name, default=None):
        try:
            pass
        except:
            traceback.print_exc()
        return default
    #
    # #
    def get_symbol_properties(self, symbol):
        ''''''
        raise NotImplemented

    @staticmethod
    def compile(script, filename='<inline>'):
        '''编译脚本'''
        # return compile_restricted(script, filename, 'exec')
        return compile_restricted(script, filename, 'exec', policy=EARestrictingNodeTransformer)

    def do_tick(self,):
        if not self.byte_code:
            #load libs
            for lib in self.script_libs:
                lib_bc = EABase.compile(lib['code'], lib['path'])
                exec(lib_bc, self.safe_globals)
            #
            if self.script_path is None:
                self.byte_code = EABase.compile(self.script)
            else:
                # self.byte_code = EABase.compile(open(self.script_path).read(), self.script_path)
                self.byte_code = EABase.compile(self.script, self.script_path)

        self.order_log = []
        stime = time.time()
        exec(self.byte_code, self.safe_globals)
        self.last_exec_time = time.time() - stime
    #
    # def do_tick(self,):
    #     if not self.byte_code:
    #         if self.script_path is None:
    #             self.byte_code = EABase.compile(self.script)
    #         else:
    #             # self.byte_code = EABase.compile(open(self.script_path).read(), self.script_path)
    #             self.byte_code = EABase.compile(self.script, self.script_path)
    #
    #     self.order_log = []
    #     stime = time.time()
    #     exec(self.byte_code, self.safe_globals)
    #     self.last_exec_time = time.time() - stime
    #     #




from datetime import datetime

class Account(object):
    '''Account class'''
    def __init__(self, params={}):
        self.__account_dict__ = params.copy()

    @property
    def balance(self) -> float:
        return float(self.__account_dict__["balance"])

    @property
    def equity(self) -> float:
        return float(self.__account_dict__["equity"])

    @property
    def margin(self) -> float:
        return float(self.__account_dict__["margin"])

    @property
    def free_margin(self) -> float:
        return float(self.__account_dict__["free_margin"])

    @property
    def credit(self) -> float:
        return float(self.__account_dict__["credit"])

    @property
    def profit(self) -> float:
        return float(self.__account_dict__["profit"])

    @property
    def margin_level(self) -> float:
        return float(self.__account_dict__["margin_level"])

    @property
    def leverage(self) -> int:
        return int(self.__account_dict__["leverage"])

    @property
    def currency(self) -> str:
        return str(self.__account_dict__["currency"])

    @property
    def free_margin_mode(self) -> float:
        return float(self.__account_dict__["free_margin_mode"])

    @property
    def stop_out_level(self) -> float:
        return float(self.__account_dict__["stop_out_level"])

    @property
    def stop_out_mode(self) -> float:
        return float(self.__account_dict__["stop_out_mode"])

    @property
    def company(self) -> str:
        return str(self.__account_dict__["company"])

    @property
    def name(self) -> str:
        return str(self.__account_dict__["name"])

    @property
    def number(self) -> str:
        return str(self.__account_dict__["number"])

    @property
    def server(self) -> str:
        return str(self.__account_dict__["server"])

    @property
    def trade_mode(self) -> int:
        return int(self.__account_dict__["trade_mode"])

    @property
    def limit_orders(self) -> int:
        return int(self.__account_dict__["limit_orders"])

    @property
    def margin_so_mode(self) -> int:
        return int(self.__account_dict__["margin_so_mode"])

    @property
    def trade_allowed(self) -> int:
        return int(self.__account_dict__["trade_allowed"])

    @property
    def trade_expert(self) -> int:
        return int(self.__account_dict__["trade_expert"])

    @property
    def margin_so_call(self) -> float:
        return float(self.__account_dict__["margin_so_call"])

    @property
    def margin_so_so(self) -> float:
        return float(self.__account_dict__["margin_so_so"])

    @property
    def commission(self) -> float:
        return float(self.__account_dict__.get("commission", 0.0))


class AccountDataNumber(object):
    def __init__(self, account_data, data, timeframe, getitem_callback):
        self.__timeframe__ = timeframe
        self.__data__ = data
        self.__account_data__ = account_data
        self.__getitem_callback__ = getitem_callback

    def __getitem__(self, key):
        return self.__getitem_callback__(self.__data__, self.__timeframe__,
                                         key)

class AccountDataString(object):
    def __init__(self, account_data, data, timeframe, getitem_callback):
        self.__timeframe__ = timeframe
        self.__data__ = data
        self.__account_data__ = account_data
        self.__getitem_callback__ = getitem_callback

    def __getitem__(self, key):
        return self.__getitem_callback__(self.__data__, self.__timeframe__,
                                         key)


class AccountDataTime(object):
    def __init__(self, account_data, data, timeframe, getitem_callback):
        self.__timeframe__ = timeframe
        self.__time__ = data
        self.__account_data__ = account_data
        self.__getitem_callback__ = getitem_callback

    def __getitem__(self, key):
        ts = self.__getitem_callback__(self.__time__, self.__timeframe__,
                                       key)
        if ts is None:
            return None
        # return datetime.fromtimestamp(ts)
        return datetime.utcfromtimestamp(ts)

class AccountData(object):
    def __init__(self, data, timeframe, getitem_callback):
        self.__timeframe__ = timeframe
        self.__data__ = data
        self.__size__ = data.size
        self.__time__ = AccountDataTime(self, data['t'], timeframe, getitem_callback=getitem_callback)
        #
        self.__credit__ = AccountDataNumber(self, data['credit'], timeframe, getitem_callback=getitem_callback)
        self.__balance__ = AccountDataNumber(self, data['balance'], timeframe, getitem_callback=getitem_callback)
        self.__max_balance__ = AccountDataNumber(self, data['max_balance'], timeframe, getitem_callback=getitem_callback)
        self.__min_balance__ = AccountDataNumber(self, data['min_balance'], timeframe, getitem_callback=getitem_callback)
        self.__equity__ = AccountDataNumber(self, data['equity'], timeframe, getitem_callback=getitem_callback)
        self.__max_equity__ = AccountDataNumber(self, data['max_equity'], timeframe, getitem_callback=getitem_callback)
        self.__min_equity__ = AccountDataNumber(self, data['min_equity'], timeframe, getitem_callback=getitem_callback)
        self.__margin__ = AccountDataNumber(self, data['margin'], timeframe, getitem_callback=getitem_callback)
        self.__max_margin__ = AccountDataNumber(self, data['max_margin'], timeframe, getitem_callback=getitem_callback)
        self.__min_margin__ = AccountDataNumber(self, data['min_margin'], timeframe, getitem_callback=getitem_callback)
        self.__margin_level__ = AccountDataNumber(self, data['margin_level'], timeframe, getitem_callback=getitem_callback)
        self.__profit__ = AccountDataNumber(self, data['profit'], timeframe, getitem_callback=getitem_callback)
        self.__max_profit__ = AccountDataNumber(self, data['max_profit'], timeframe, getitem_callback=getitem_callback)
        self.__min_profit__ = AccountDataNumber(self, data['min_profit'], timeframe, getitem_callback=getitem_callback)
        self.__free_margin__ = AccountDataNumber(self, data['free_margin'], timeframe, getitem_callback=getitem_callback)
        self.__commission__ = AccountDataNumber(self, data['commission'], timeframe, getitem_callback=getitem_callback)
        #static
        self.__currency__ = AccountDataString(self, data['currency'], timeframe, getitem_callback=getitem_callback)
        self.__free_margin_mode__ = AccountDataNumber(self, data['free_margin_mode'], timeframe, getitem_callback=getitem_callback)
        self.__leverage__ = AccountDataNumber(self, data['leverage'], timeframe, getitem_callback=getitem_callback)
        self.__stop_out_level__ = AccountDataNumber(self, data['stop_out_level'], timeframe, getitem_callback=getitem_callback)
        self.__stop_out_mode__ = AccountDataNumber(self, data['stop_out_mode'], timeframe, getitem_callback=getitem_callback)
        self.__company__ = AccountDataString(self, data['company'], timeframe, getitem_callback=getitem_callback)
        self.__name__ = AccountDataString(self, data['name'], timeframe, getitem_callback=getitem_callback)
        self.__number__ = AccountDataString(self, data['number'], timeframe, getitem_callback=getitem_callback)
        self.__server__ = AccountDataString(self, data['server'], timeframe, getitem_callback=getitem_callback)
        self.__trade_mode__ = AccountDataNumber(self, data['trade_mode'], timeframe, getitem_callback=getitem_callback)
        self.__limit_orders__ = AccountDataNumber(self, data['limit_orders'], timeframe, getitem_callback=getitem_callback)
        self.__margin_so_mode__ = AccountDataNumber(self, data['margin_so_mode'], timeframe, getitem_callback=getitem_callback)
        self.__trade_allowed__ = AccountDataNumber(self, data['trade_allowed'], timeframe, getitem_callback=getitem_callback)
        self.__trade_expert__ = AccountDataNumber(self, data['trade_expert'], timeframe, getitem_callback=getitem_callback)
        self.__margin_so_call__ = AccountDataNumber(self, data['margin_so_call'], timeframe, getitem_callback=getitem_callback)
        self.__margin_so_so__ = AccountDataNumber(self, data['margin_so_so'], timeframe, getitem_callback=getitem_callback)
        self.__getitem_callback__ = getitem_callback

    def __getitem__(self, key):
        return self.__getitem_callback__(self.__data__, self.__timeframe__, key)

    @property
    def size(self):
        return self.__size__

    @property
    def commission(self):
        return self.__commission__

    @property
    def credit(self):
        return self.__credit__

    @property
    def balance(self):
        return self.__balance__

    @property
    def max_balance(self):
        return self.__max_balance__

    @property
    def min_balance(self):
        return self.__min_balance__

    @property
    def equity(self):
        return self.__equity__

    @property
    def max_equity(self):
        return self.__max_equity__

    @property
    def min_equity(self):
        return self.__min_equity__

    @property
    def margin(self):
        return self.__margin__

    @property
    def max_margin(self):
        return self.__max_margin__

    @property
    def min_margin(self):
        return self.__min_margin__

    @property
    def margin_level(self):
        return self.__margin_level__

    @property
    def profit(self):
        return self.__profit__

    @property
    def max_profit(self):
        return self.__max_profit__

    @property
    def min_profit(self):
        return self.__min_profit__

    @property
    def free_margin(self):
        return self.__free_margin__

    @property
    def currency(self):
        return self.__currency__

    @property
    def free_margin_mode(self):
        return self.__free_margin_mode__

    @property
    def leverage(self):
        return self.__leverage__

    @property
    def stop_out_level(self):
        return self.__stop_out_level__

    @property
    def stop_out_mode(self):
        return self.__stop_out_mode__

    @property
    def company(self):
        return self.__company__

    @property
    def name(self):
        return self.__name__

    @property
    def number(self):
        return self.__number__

    @property
    def server(self):
        return self.__server__

    @property
    def trade_mode(self):
        return self.__trade_mode__

    @property
    def limit_orders(self):
        return self.__limit_orders__

    @property
    def margin_so_mode(self):
        return self.__margin_so_mode__

    @property
    def trade_allowed(self):
        return self.__trade_allowed__

    @property
    def trade_expert(self):
        return self.__trade_expert__

    @property
    def margin_so_call(self):
        return self.__margin_so_call__

    @property
    def margin_so_so(self):
        return self.__margin_so_so__


#
print(f"time={Time()}")
point = SymbolInfo("point")
volume = 0.01
errid, result = Buy(volume=volume)
order_uid = result['order_uid']
print(f"Buy: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)


#Buy with sl and tp
stop_loss=Bid()-15*point
take_profit=Ask()+30*point
errid, result = Buy(volume=0.01, price=Ask(), stop_loss=stop_loss,
                    take_profit=take_profit)
order_uid = result['order_uid']
print(f"Buy: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

#Buy all params
stop_loss=Bid()-15*point
take_profit=Ask()+30*point
magic_number=4291651
errid, result = Buy(volume=0.01, price=Ask(), stop_loss=stop_loss,
                    take_profit=take_profit, magic_number=magic_number,
                    symbol=Symbol(), slippage=3, arrow_color="white")
order_uid = result['order_uid']
print(f"Buy: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

#
stop_loss = stop_loss - 15 * point
take_profit = take_profit + 30 * point
errid, result = ModifyOrder(order_uid, stop_loss=stop_loss, take_profit=take_profit)
order_uid = result['order_uid']
print(f"ModifyOrder: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

#close order
errid, result = CloseOrder(order_uid)
order_uid = result['order_uid']
print(f"CloseOrder: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)



# Sell
errid, result = Sell(volume=0.01)
order_uid = result['order_uid']
print(f"Sell: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

#Sell with sl and tp
stop_loss = Ask() + 15 * point
take_profit = Bid() - 30 * point
errid, result = Sell(volume=0.01, price=Bid(), stop_loss=Ask()+15*point,
                    take_profit=Bid()-30*point)
order_uid = result['order_uid']
print(f"Sell: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

#Sell all params
magic_number=4291652
errid, result = Sell(volume=volume, price=Bid(), stop_loss=stop_loss,
                    take_profit=take_profit, magic_number=magic_number,
                    symbol=Symbol(), slippage=3, arrow_color="red")
order_uid = result['order_uid']
print(f"Sell: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)
#
stop_loss = stop_loss + 15 * point
take_profit = take_profit - 30 * point
errid, result = ModifyOrder(order_uid, stop_loss=stop_loss, take_profit=take_profit)
order_uid = result['order_uid']
print(f"ModifyOrder: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)
#close order
errid, result = CloseOrder(order_uid)
order_uid = result['order_uid']
print(f"CloseOrder: errid={errid}, order_uid={order_uid}")
order = GetOrder(order_uid)

StopTester()


import json
from bson import json_util
import uuid
import hashlib
from hashids import Hashids
from datetime import datetime, date
from dateutil.parser import parse
from pixiu.api.defines import OrderCommand
# ----------------------------------------------------------------------------------------------------------------------
#    UTILS
# ----------------------------------------------------------------------------------------------------------------------
def order_is_long(order_cmd):
    long_types = [OrderCommand.BUY, OrderCommand.BUYLIMIT, OrderCommand.BUYSTOP]
    return order_cmd.upper() in long_types

def order_is_short(order_cmd):
    long_types = [OrderCommand.SELL, OrderCommand.SELLLIMIT, OrderCommand.SELLSTOP]
    return order_cmd.upper() in long_types

def order_is_market(order_cmd) -> bool:
    market_types = [OrderCommand.BUY, OrderCommand.SELL]
    return order_cmd.upper() in market_types

def order_is_stop(order_cmd) -> bool:
    stop_types = [OrderCommand.BUYSTOP, OrderCommand.SELLSTOP]
    return order_cmd.upper() in stop_types

def order_is_limit(order_cmd) -> bool:
    limit_types = [OrderCommand.BUYLIMIT, OrderCommand.SELLLIMIT]
    return order_cmd.upper() in limit_types

def order_is_pending(order_cmd) -> bool:
    return order_is_limit(order_cmd) or order_is_stop(order_cmd)

def parse_timeframe(timeframe):
    '''Parse the timeframe'''
    unit = timeframe[0]
    idx = 1
    if unit == 'm':
        if len(timeframe) > 1 and timeframe[1] == 'n':
            unit = 'mn'
            idx = 2

    count = timeframe[idx:]
    if not count:
        count = 1
    return unit, int(count)


def calculate_minute_intervals(t_unit, t_count):
    '''Calculate minute intervals'''
    return int(calculate_second_intervals(t_unit, t_count) / 60)

def calculate_second_intervals(t_unit, t_count):
    '''Calculate second intervals'''
    seconds = None
    MINUTE_SEC = 60
    HOUR_SEC = 3600
    DAY_SEC = 86400
    WEEK_SEC = 604800
    if t_unit == 's':
        seconds = t_count
    elif t_unit == 'm':
        seconds = MINUTE_SEC * t_count
    elif t_unit == 'h':
        seconds = HOUR_SEC * t_count
    elif t_unit == 'd':
        seconds = DAY_SEC * t_count
    elif t_unit == 'w':
        seconds = WEEK_SEC * t_count
    else:
        pass
    return seconds

def timeframe_to_seconds(timeframe):
    '''Timeframe to seconds'''
    t_unit, t_count = parse_timeframe(timeframe)
    return calculate_second_intervals(t_unit, t_count)

def parse_datetime_string(time_str, ignoretz=True):
    ''''''
    if time_str is None:
        return None
    if(isinstance(time_str, datetime) or isinstance(time_str, date)):
        return time_str
    return parse(time_str, ignoretz=ignoretz)

def load_json(source, object_hook=json_util.object_hook):
    """"""
    ret = source
    if isinstance(source, bytes):
        source = source.decode('utf-8')
    if isinstance(source, str):
        ret = json.loads(source, object_hook=object_hook)
    return ret

def dump_json(value, default=json_util.default):
    """"""
    svalue = json.dumps(value, default=default)
    return svalue

def uuid_str(short=True):
    hashids = Hashids()
    key = hashlib.md5(str(uuid.uuid4()).encode("utf-8")).hexdigest()
    if short:
        key = hashids.encode(int(key, 16))
    return key



from ..base.api_base import (API_V1_Base)
import numpy as np
import talib
import logging
from datetime import datetime
from pixiu.api import RunModeValue
from pixiu.api.v1 import (OrderCommand, OrderType, IndicatiorID, SymbolIndicator, SymbolPrice, SymbolData,
                      Order, Account, Symbol, AccountData, ErrorID, OrderResult, DataScope)
log = logging.getLogger(__name__)


#---------------------------------------------------------------------------------------------------------------------
# TesterAPI_V1
#---------------------------------------------------------------------------------------------------------------------
class TesterAPI_V1(API_V1_Base):
    def __init__(self, tester, data_source, default_symbol):
        super(TesterAPI_V1, self).__init__(data_source, default_symbol)
        self.tester = tester

    def _print_(self, *args, **kwargs):
        """"""
        return self.tester.get_print_factory(*args, **kwargs)
    #
    def Buy(self, volume, type=OrderType.MARKET, price=0, stop_loss=None, take_profit=None, comment=None,
            magic_number=None, symbol=None, slippage=None, arrow_color=None, expiration=None, tags=None):
        if symbol is None:
            symbol = self.default_symbol
        if price <= 0:
            price = self.Ask()

        if type == OrderType.MARKET:
            oct = OrderCommand.BUY
        elif type == OrderType.LIMIT:
            oct = OrderCommand.BUYLIMIT
        elif type == OrderType.STOP:
            oct = OrderCommand.BUYSTOP
        else:
            raise NotImplementedError

        return self.tester.open_order(symbol, oct, price, volume, stop_loss, take_profit,
                                           comment=comment, magic_number=magic_number, slippage=slippage,
                                           arrow_color=arrow_color, tags=tags)

    def Sell(self, volume, type=OrderType.MARKET, price=0, stop_loss=None, take_profit=None, comment=None,
             magic_number=None, symbol=None, slippage=None, arrow_color=None, expiration=None, tags=None):
        if symbol is None:
            symbol = self.default_symbol
        if price <= 0:
            price = self.Bid()

        if type == OrderType.MARKET:
            oct = OrderCommand.SELL
        elif type == OrderType.LIMIT:
            oct = OrderCommand.SELLLIMIT
        elif type == OrderType.STOP:
            oct = OrderCommand.SELLSTOP
        else:
            raise NotImplementedError
        return self.tester.open_order(symbol, oct, price, volume, stop_loss, take_profit,
                                           comment=comment, magic_number=magic_number, slippage=slippage,
                                           arrow_color=arrow_color, tags=tags)

    def ModifyOrder(self, uid, price=None, stop_loss=None, take_profit=None, comment=None,
                     arrow_color=None, expiration=None, tags=None):
        return self.tester.modify_order(uid, price, stop_loss, take_profit,
                              comment=comment, arrow_color=arrow_color, expiration=expiration, tags=tags)

    #
    def CloseOrder(self, uid, volume: float=None, price=None,  slippage=None, arrow_color=None, tags=None):
        return self.tester.close_order(uid, volume, price, slippage=slippage,
                              arrow_color=arrow_color, tags=tags)

    def CloseMultiOrders(self, orders):
        return self.tester.close_multi_orders(orders)

    def Notify(self, message):
        return self.tester.notify(message)

    def RunMode(self):
        return RunModeValue.TEST

    def WaitCommand(self, uid, timeout=120):
        return self.tester.wait_command(uid, timeout)

    def AcquireLock(self, name, timeout=60):
        return self.tester.acquire_lock(name, timeout)

    def ReleaseLock(self, name):
        return self.tester.release_lock(name)

    def DefaultTimeFrame(self):
        return self.tester.tick_timeframe

    def GetSymbolData(self, symbol: str, timeframe: str, size: int):
        data = self.tester.get_data_info(symbol, timeframe)
        return SymbolData(data, timeframe, getitem_callback=self.tester.symbol_datagetitem_callback,
                          getitem_index=self.tester.symbol_datagetitem_index)

    def GetAccount(self):
        return Account(self.tester.get_account())

    def GetSymbol(self, symbol=None):
        if symbol is None:
            symbol = self.default_symbol
        sp = self.tester.get_symbol_properties(symbol)
        return Symbol(sp)

    def AccountEquity(self):
        return self.GetAccount().equity

    def AccountFreeMargin(self):
        return self.GetAccount().free_margin

    def SymbolInfo(self, item, symbol=None, default=None):
        if symbol is None:
            symbol = self.default_symbol
        sp = self.tester.get_symbol_properties(symbol)
        return sp.get(item, default)

    def OrderStats(self, order_uids):
        """"""
        ret = {}
        ret['count'] = len(order_uids)
        ret['volume'] = 0
        ret['profit'] = 0
        ret['open_highest'] = 0
        ret['open_lowest'] = 99999999
        #nHoldCount, nTotalVolume, nTotalProfit, nHightest, nLowest
        for oid in order_uids:
            order = self.GetOrder(oid)
            if order is None:
                continue
            ret['volume'] += order.volume
            ret['profit'] += order.profit
            if order.open_price > ret['open_highest']:
                ret['open_highest'] = order.open_price
            if order.open_price < ret['open_lowest']:
                ret['open_lowest'] = order.open_price
        return ret

    def Plot(self, series):
        self.tester.plot('default', series)

    def GetParam(self, name, default=None):
        '''Get params'''
        return self.tester.get_param(name, default)

    def GetSettings(self, name, default=None):
        return self.tester.get_settings(name, default)

    def DeleteData(self, name, scope: int = DataScope.EA):
        return self.tester.delete_data(name, scope)

    def LoadData(self, name, scope: int = DataScope.EA, format='json'):
        return self.tester.load_data(name, scope, format)

    def SaveData(self, name, data, scope: int = DataScope.EA, format='json'):
        return self.tester.save_data(name, data, scope, format)

    def GetOrder(self, order_uid: str):
        return Order(self.tester.get_order(order_uid=order_uid))

    def Close(self, shift=0, symbol=None) -> float:
        return self.tester.Close(shift=shift, symbol=symbol)

    def Open(self, shift=0, symbol=None) -> float:
        return self.tester.Open(shift=shift, symbol=symbol)

    def High(self, shift=0, symbol=None) -> float:
        return self.tester.High(shift=shift, symbol=symbol)

    def Low(self, shift=0, symbol=None) -> float:
        return self.tester.Low(shift=shift, symbol=symbol)

    def Ask(self, shift=0, symbol=None) -> float:
        return self.tester.Ask(shift=shift, symbol=symbol)

    def Bid(self, shift=0, symbol=None) -> float:
        return self.tester.Bid(shift=shift, symbol=symbol)

    def Time(self, shift=0, symbol=None) -> datetime:
        return self.tester.Time(shift=shift, symbol=symbol)

    def Volume(self, shift=0, symbol=None) -> float:
        return self.tester.Volume(shift=shift, symbol=symbol)

    def Symbol(self, shift=0) -> str:
        return self.tester.symbol

    def GetOpenedOrderUIDs(self, symbol: str = None, scope: int = DataScope.EA):
        ret = self.tester.get_order_dict(symbol, status="opened", scope=scope)
        return list(ret.keys())

    def GetPendingOrderUIDs(self, symbol: str = None, scope: int = DataScope.EA):
        ret = self.tester.get_order_dict(symbol, status="pending", scope=scope)
        return list(ret.keys())

    def GetClosedOrderUIDs(self, symbol: str = None, scope: int = DataScope.EA):
        ret = self.tester.get_order_dict(symbol, status="closed", scope=scope)
        return list(ret.keys())

    def StopTester(self, code: int=0, message: str=None):
        return self.tester.stop_tester(code, message)

    def __calculate_indicator__(self, indicator_id, price_data, *args, **kwargs):
        #indicator_id, price_data, period, ma_type
        if not isinstance(price_data, SymbolPrice) and not isinstance(price_data, SymbolData):
            return None
        if indicator_id == IndicatiorID.MA:
            ret = talib.MA(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.AD:
            #high, low, close, volume
            ret = talib.AD(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                           price_data.volume.__price__)
        elif indicator_id == IndicatiorID.ADX:
            #high, low, close[, timeperiod=?]
            ret = talib.ADX(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                            *args, **kwargs)
        elif indicator_id == IndicatiorID.ATR:
            #(high, low, close[, timeperiod=?]
            ret = talib.ATR(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                            *args, **kwargs)
        elif indicator_id == IndicatiorID.BANDS:
            # real[, timeperiod=?, nbdevup=?, nbdevdn=?, matype=?]
            # nbDevUp
            # Deviation multiplier for upper band. Valid range from TRADER_REAL_MIN to TRADER_REAL_MAX.
            #
            # nbDevDn
            # Deviation multiplier for lower band. Valid range from TRADER_REAL_MIN to TRADER_REAL_MAX.
            upper, middle, lower = talib.BBANDS(price_data.__price__, *args, **kwargs)
            ret = np.array([upper, middle, lower]).T #or: np.transpose(a), a.transpose(
            # ret = []
            # for idx in range(upper.volume):
            #     ret.append((upper[idx], middle[idx], lower[idx]))
        elif indicator_id == IndicatiorID.CCI:
            # high, low, close[, timeperiod=?]
            ret = talib.CCI(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                            *args, **kwargs)
        elif indicator_id == IndicatiorID.CHAIKIN:
            # high, low, close, volume[, fastperiod=?, slowperiod=?]
            ret = talib.ADOSC(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                              price_data.volume.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.DEMA:
            #DEMA(real[, timeperiod=?])
            ret = talib.DEMA(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.MOMENTUM:
            # MOM(real[, timeperiod=?])
            ret = talib.MOM(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.MFI:
            # MFI(high, low, close, volume[, timeperiod=?])
            ret = talib.MFI(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                              price_data.volume.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.MACD:
            # MACD(real[, fastperiod=?, slowperiod=?, signalperiod=?])
            macd, macdsignal, macdhist = talib.MACD(price_data.__price__, *args, **kwargs)
            ret = np.array([macd, macdsignal, macdhist]).T #or: np.transpose(a), a.transpose(
        elif indicator_id == IndicatiorID.OBV:
            #  OBV(real, volume)
            ret = talib.OBV(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.SAR:
            # SAR(high, low[, acceleration=?, maximum=?])
            ret = talib.SAR(price_data.high.__price__, price_data.low.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.RSI:
            # RSI(real[, timeperiod=?])
            ret = talib.RSI(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.STDDEV:
            # STDDEV(real[, timeperiod=?, nbdev=?])
            ret = talib.STDDEV(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.STOCHASTIC:
            #  STOCH(high, low, close[, fastk_period=?, slowk_period=?, slowk_matype=?, slowd_period=?, slowd_matype=?])
            slowk, slowd = talib.STOCH(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                              *args, **kwargs)
            ret = np.array([slowk, slowd]).T #or: np.transpose(a), a.transpose(
        elif indicator_id == IndicatiorID.TEMA:
            # TEMA(real[, timeperiod=?])
            ret = talib.TEMA(price_data.__price__, *args, **kwargs)
        elif indicator_id == IndicatiorID.WPR:
            #  WILLR(high, low, close[, timeperiod=?])
            ret = talib.WILLR(price_data.high.__price__, price_data.low.__price__, price_data.close.__price__,
                              *args, **kwargs)
        else:
            return None

        return ret

    def __get_indicator__(self, cache_name, shift, indicator_id, price_data, *args, **kwargs):
        ret = price_data.indicators.get(cache_name, None)
        if ret is None:
            #data, ts_index, timeframe, getitem_callback
            ret = SymbolIndicator(self.__calculate_indicator__(indicator_id, price_data, *args, **kwargs),
                                  price_data.ts_index,
                                  price_data.timeframe,
                                  getitem_callback=price_data.__getitem_callback__,
                                  getitem_index=price_data.__getitem_index__,
                                  )
            price_data.indicators[cache_name] = ret
        if ret is not None and shift is not None:
            return ret[shift]
        return ret

    def iMA(self, price_data, timeperiod, matype, shift=0):
        return self.__get_indicator__(f"ma:{matype}:{timeperiod}", shift, IndicatiorID.MA, price_data, timeperiod, matype)

    def iAD(self, symbol_data, shift=0):
        return self.__get_indicator__(f"ad:", shift, IndicatiorID.AD, symbol_data)

    def iADX(self, symbol_data, timeperiod, shift=0):
        '''
        Average Directional Movement Index
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"adx:{timeperiod}", shift, IndicatiorID.ADX, symbol_data, timeperiod)

    def iATR(self, symbol_data, timeperiod, shift=0):
        '''
        Average True Range
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"atr:{timeperiod}", shift, IndicatiorID.ATR, symbol_data, timeperiod)


    def iBands(self, symbol_data, timeperiod, nbdevup, nbdevdn, matype, shift=0):
        '''
        Bollinger Bands®
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"bands:{matype}:{timeperiod}:{nbdevup}:{nbdevdn}", shift, IndicatiorID.BANDS, symbol_data, timeperiod, nbdevup,
                                      nbdevdn, matype)


    def iCCI(self, symbol_data, timeperiod, shift=0):
        '''
        Commodity Channel Index
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"cci:{timeperiod}", shift, IndicatiorID.CCI, symbol_data, timeperiod)


    def iChaikin(self, symbol_data, fastperiod, slowperiod, shift=0):
        '''
        Chaikin Oscillator/ ADOSC                Chaikin A/D Oscillator
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"chaikin:{fastperiod}:{slowperiod}", shift, IndicatiorID.CHAIKIN, symbol_data, fastperiod, slowperiod)


    def iDEMA(self, symbol_data, timeperiod, shift=0):
        '''
        Double Exponential Moving Average
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"dema:{timeperiod}", shift, IndicatiorID.DEMA, symbol_data, timeperiod)


    def iMomentum(self, symbol_data, timeperiod, shift=0):
        '''
        Momentum / MOM                  Momentum
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"momentum:{timeperiod}", shift, IndicatiorID.MOMENTUM, symbol_data, timeperiod)


    def iMFI(self, symbol_data, timeperiod, shift=0):
        '''
        Money Flow Index
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"mfi:{timeperiod}", shift, IndicatiorID.MFI, symbol_data, timeperiod)


    def iMACD(self, symbol_data, fastperiod, slowperiod, signalperiod, shift=0):
        '''
        Moving Averages Convergence-Divergence
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"macd:{fastperiod}:{slowperiod}:{signalperiod}", shift, IndicatiorID.MACD, symbol_data, fastperiod, slowperiod,
                                      signalperiod)


    def iOBV(self, symbol_data, shift=0):
        '''
        On Balance Volume
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"obv:", shift, IndicatiorID.OBV, symbol_data,
                                      symbol_data.__symbol_data__.volume.__price__)

    def iSAR(self, symbol_data, acceleration, maximum, shift=0):
        '''
        Parabolic Stop And Reverse System
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"sar:{acceleration}:{maximum}", shift, IndicatiorID.SAR, symbol_data, acceleration, maximum)


    def iRSI(self, symbol_data, timeperiod, shift=0):
        '''
        Relative Strength Index
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"rsi:{timeperiod}", shift, IndicatiorID.RSI, symbol_data, timeperiod)

    def iStdDev(self, symbol_data, timeperiod, nbdev, shift=0):
        '''
        Standard Deviation
        Parameters
            symbol_data:
        '''
        return self.__get_indicator__(f"stddev:{timeperiod}:{nbdev}", shift, IndicatiorID.STDDEV, symbol_data, timeperiod, nbdev)

    def iStochastic(self, symbol_data, fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype, shift=0):
        '''
        Stochastic Oscillator
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"stochastic:{fastk_period}:{slowk_period}:{slowk_matype}:{slowd_period}:{slowd_matype}", shift, IndicatiorID.STOCHASTIC, symbol_data,
                                      fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype)

    def iTEMA(self, symbol_data, timeperiod, shift=0):
        '''
        Triple Exponential Moving Average
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"tema:{timeperiod}", shift, IndicatiorID.TEMA, symbol_data, timeperiod)

    def iWPR(self, symbol_data, timeperiod, shift=0):
        '''
        Williams' Percent Range
        Parameters
            symbol_data:
        '''
        # real = ADX(high, low, close, timeperiod=14)
        return self.__get_indicator__(f"wpr:{timeperiod}", shift, IndicatiorID.WPR, symbol_data, timeperiod)




from datetime import datetime

import numpy as np
import pandas
from pixiu.api.utils import (parse_timeframe, calculate_minute_intervals, timeframe_to_seconds)


class Symbol(object):
    '''Symbol class'''
    def __init__(self, params={}):
        self.__symbol_dict__ = params.copy()

    @property
    def type(self) -> int:
        return int(self.__symbol_dict__["type"])

    @property
    def name(self) -> str:
        return str(self.__symbol_dict__["symbol"])

    @property
    def spread(self) -> int:
        return int(self.__symbol_dict__["spread"])

    @property
    def digits(self) -> int:
        return int(self.__symbol_dict__["digits"])

    @property
    def stop_level(self) -> int:
        return int(self.__symbol_dict__.get("stop_level", 0))

    @property
    def volume_min(self) -> float:
        return float(self.__symbol_dict__["volume_min"])

    @property
    def trade_contract_size(self) -> float:
        return float(self.__symbol_dict__["trade_contract_size"])

    @property
    def point(self) -> float:
        return float(self.__symbol_dict__["point"])

    @property
    def currency_profit(self) -> str:
        return str(self.__symbol_dict__["currency_profit"])

    @property
    def currency_base(self) -> str:
        return str(self.__symbol_dict__["currency_base"])

    @property
    def currency_margin(self) -> str:
        return str(self.__symbol_dict__["currency_margin"])


#---------------------------------------------------------------------------------------------------------------------
# SymbolData
#---------------------------------------------------------------------------------------------------------------------
class SymbolIndicator(object):
    def __init__(self, data, ts_index, timeframe, getitem_callback, getitem_index):
        self.__timeframe__ = timeframe
        self.__timeframe_seconds__ = timeframe_to_seconds(timeframe)
        self.data = data
        self.__ts_index__ = ts_index
        self.__getitem_callback__ = getitem_callback
        self.__getitem_index__ = getitem_index

    @property
    def ts_index(self):
        return self.__ts_index__

    @property
    def timeframe(self):
        return self.__timeframe__

    @property
    def timeframe_seconds(self):
        return self.__timeframe_seconds__

    def __getitem__(self, key):
        # return self.__getitem_callback__(self.data, self.__timeframe__, key)
        return self.__getitem_callback__(self.data, self.__ts_index__,
                                       self.__timeframe__,
                                       self.__timeframe_seconds__,
                                       key, fail_value=np.NaN)


class SymbolPrice(object):
    def __init__(self, symbol_data, price, getitem_callback, getitem_index):
        # self.__timeframe__ = timeframe
        # self.__timeframe_seconds__ = timeframe_to_seconds(timeframe)
        self.__price__ = price
        self.__symbol_data__ = symbol_data
        self.indicators = {}
        self.__getitem_callback__ = getitem_callback
        self.__getitem_index__ = getitem_index

    @property
    def ts_index(self):
        return self.__symbol_data__.ts_index

    @property
    def timeframe(self):
        return self.__symbol_data__.timeframe

    @property
    def timeframe_seconds(self):
        return self.__symbol_data__.timeframe_seconds

    def __getitem__(self, key):
        ret = self.__getitem_callback__(self.__price__, self.__symbol_data__.ts_index,
                                       self.__symbol_data__.timeframe,
                                       self.__symbol_data__.timeframe_seconds,
                                       key, fail_value=np.NaN)
        return ret

    def __add__(self, other):
        return self.__price__ + other.__price__

    def __sub__(self, other):
        return self.__price__ - other.__price__

    def __mul__(self, other):
        return self.__price__ * other.__price__

    def __truediv__(self, other):
        return self.__price__ / other.__price__

    def __floordiv__(self, other):
        return self.__price__ // other.__price__

    def __mod__(self, other):
        return self.__price__ % other.__price__

    def __pow__(self, other):
        return self.__price__ ** other.__price__

    def __lt__(self, other):
        return self.__price__ < other.__price__

    def __gt__(self, other):
        return self.__price__ > other.__price__

    def __le__(self, other):
        return self.__price__ <= other.__price__

    def __ge__(self, other):
        return self.__price__ >= other.__price__

    def __eq__(self, other):
        return self.__price__ == other.__price__

    def __ne__(self, other):
        return self.__price__ != other.__price__


class SymbolTime(object):
    def __init__(self, symbol_data, price, getitem_callback, getitem_index):
        self.__time__ = price
        self.__symbol_data__ = symbol_data
        self.indicators = {}
        self.__getitem_callback__ = getitem_callback
        self.__getitem_index__ = getitem_index

    @property
    def ts_index(self):
        return self.__symbol_data__.ts_index

    @property
    def timeframe(self):
        return self.__symbol_data__.timeframe

    @property
    def timeframe_seconds(self):
        return self.__symbol_data__.timeframe_seconds

    def __getitem__(self, key):
        ts = self.__getitem_callback__(self.__time__, self.__symbol_data__.ts_index,
                                       self.__symbol_data__.timeframe,
                                       self.__symbol_data__.timeframe_seconds,
                                       key, fail_value=None)
        if ts is None:
            return None
        # return datetime.fromtimestamp(ts)

        return datetime.utcfromtimestamp(ts)

class SymbolData(object):
    def __init__(self, data, timeframe, getitem_callback, getitem_index):
        self.__timeframe__ = timeframe
        self.__timeframe_seconds__ = timeframe_to_seconds(timeframe)
        self.__data__ = data
        self.__size__ = data.size
        self.__ts_index__ = data['t']
        self.__time__ = SymbolTime(self, data['t'], getitem_callback=getitem_callback, getitem_index=getitem_index)
        self.__close_price__ = SymbolPrice(self, data['c'], getitem_callback=getitem_callback,
                                           getitem_index=getitem_index)
        self.__open_price__ = SymbolPrice(self, data['o'], getitem_callback=getitem_callback,
                                          getitem_index=getitem_index)
        self.__high_price__ = SymbolPrice(self, data['h'], getitem_callback=getitem_callback,
                                          getitem_index=getitem_index)
        self.__low_price__ = SymbolPrice(self, data['l'], getitem_callback=getitem_callback,
                                         getitem_index=getitem_index)
        self.__ask_price__ = SymbolPrice(self, data['a'], getitem_callback=getitem_callback,
                                         getitem_index=getitem_index)
        self.__bid_price__ = SymbolPrice(self, data['b'], getitem_callback=getitem_callback,
                                         getitem_index=getitem_index)
        self.__volume__ = SymbolPrice(self, data['v'], getitem_callback=getitem_callback,
                                      getitem_index=getitem_index)
        self.indicators = {}
        self.__getitem_callback__ = getitem_callback
        self.__getitem_index__ = getitem_index

    def __getitem__(self, key):
        # return self.__getitem_callback__(self.__data__, key)
        return self.__getitem_callback__(self.__data__, self.__data__['t'],
                                       self.timeframe,
                                       self.timeframe_seconds,
                                       key)

    def to_dataframe(self, index=0, size=None):
        if index < 0:
            return None
        if size is None or size > self.__size__:
            size = self.__size__
        offset = self.__getitem_index__(self.__data__['t'],
                                       self.timeframe,
                                       self.timeframe_seconds,
                                       index)
        if offset >= 0:
            idx = offset - size - 1
            a = self.__data__[0 if idx < 0 else idx: offset+1]
        else:
            a = self.__data__[self.__size__ - size : self.__size__ - offset + 1]
        a = a[::-1]
        ret = pandas.DataFrame({'time': a['t'], 'symbol': a['s'],
                                'open': a['o'], 'high': a['h'], 'low': a['l'], 'close': a['c'],
                                'volume': a['v'], 'ask': a['a'], 'bid': a['b']})
        return ret

    @property
    def timeframe(self):
        return self.__timeframe__

    @property
    def ts_index(self):
        return self.__ts_index__

    @property
    def timeframe_seconds(self):
        return self.__timeframe_seconds__

    @property
    def size(self):
        return self.__size__

    @property
    def close(self):
        return self.__close_price__

    @property
    def open(self):
        return self.__open_price__

    @property
    def high(self):
        return self.__high_price__

    @property
    def low(self):
        return self.__low_price__

    @property
    def ask(self):
        return self.__ask_price__

    @property
    def bid(self):
        return self.__bid_price__

    @property
    def time(self):
        return self.__time__

    @property
    def volume(self):
        return self.__volume__


#
POINT = SymbolInfo("point")
DIGITS = int(SymbolInfo("digits"))
volume = 0.01

assertEqual(AcquireLock("test"), True)
ReleaseLock("test")

score = 1.0
errid, result = Buy(volume=volume, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
errid, ret = WaitCommand(result['command_uid'])
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, 0)
assertEqual(order.take_profit, 0)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

if result['command_uid']:
    assertEqual(order.comment, f"cuid#{result['command_uid']}|")
else:
    assertEqual(order.comment, f"uid#{result['order_uid']}|")
# assertIsNone(order.magic_number)
assertIsNone(order.close_time)
assertIsNone(order.close_price)

#
score = score + 1
#Buy with sl and tp
stop_loss = round(Bid() - 15 * POINT, DIGITS)
take_profit = round(Ask() + 30 * POINT, DIGITS)
errid, result = Buy(volume=0.01, price=Ask(), stop_loss=stop_loss,
                    take_profit=take_profit, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertEqual(order.take_profit, take_profit)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#Buy all params
stop_loss = round(Bid() - 15 * POINT, DIGITS)
take_profit = round(Ask() + 30 * POINT, DIGITS)
magic_number = 4291651
errid, result = Buy(volume=0.01, price=Ask(), stop_loss=stop_loss,
                    take_profit=take_profit, magic_number=magic_number,
                    symbol=Symbol(), slippage=3, arrow_color="white", tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertEqual(order.take_profit, take_profit)
assertEqual(order.magic_number, magic_number)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#
stop_loss = round(stop_loss - 15 * POINT, DIGITS)
take_profit = round(take_profit + 30 * POINT, DIGITS)
errid, result = ModifyOrder(result['order_uid'], stop_loss=stop_loss, take_profit=take_profit, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertEqual(order.take_profit, take_profit)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#close order
# errid, result = CloseOrder(result['order_uid'], price=Ask(), volume=volume)
errid, result = CloseOrder(result['order_uid'], tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
order = GetOrder(result['order_uid'])
assertEqual(order.is_dirty, RunMode() == RunModeValue.LIVE)

errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertFalse(order.is_dirty)
assertEqual(order.symbol, Symbol())
assertIsNotNone(order.close_time)
assertEqual(order.close_price, Bid())
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.CLOSED)

#The order was closed ?
oo = GetOpenedOrderUIDs()
for t in oo:
    assertNotEqual(t, result['order_uid'])


# Sell
errid, result = Sell(volume=0.01, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, 0)
assertEqual(order.take_profit, 0)
assertIsNone(order.magic_number)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#Sell with sl and tp
stop_loss = round(Ask() + 15 * POINT, DIGITS)
take_profit = round(Bid() - 30 * POINT, DIGITS)
errid, result = Sell(volume=0.01, price=Bid(), stop_loss=Ask()+15*POINT,
                    take_profit=Bid()-30*POINT, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertAlmostEqual(order.take_profit, take_profit)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#Sell all params
magic_number=4291652
errid, result = Sell(volume=volume, price=Bid(), stop_loss=stop_loss,
                    take_profit=take_profit, magic_number=magic_number,
                    symbol=Symbol(), slippage=3, arrow_color="red", tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertEqual(order.take_profit, take_profit)
assertEqual(order.magic_number, magic_number)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#
stop_loss = round(stop_loss + 15 * POINT, DIGITS)
take_profit = round(take_profit - 30 * POINT, DIGITS)
errid, result = ModifyOrder(result['order_uid'], stop_loss=stop_loss, take_profit=take_profit, tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
order = GetOrder(result['order_uid'])
assertEqual(order.is_dirty, RunMode() == RunModeValue.LIVE)

errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertFalse(order.is_dirty)
assertEqual(order.symbol, Symbol())
assertEqual(order.uid, result['order_uid'])
assertEqual(order.volume, volume)
assertEqual(order.stop_loss, stop_loss)
assertEqual(order.take_profit, take_profit)
assertIsNone(order.close_time)
assertIsNone(order.close_price)
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.OPENED)

#close order
# errid, result = CloseOrder(result['order_uid'], price=Bid(), volume=volume)
errid, result = CloseOrder(result['order_uid'], tags={'score': score})
assertEqual(errid, 0)
assertNotEqual(result['order_uid'], None)
order = GetOrder(result['order_uid'])
assertEqual(order.is_dirty, RunMode() == RunModeValue.LIVE)

errid = exec_command()
assertEqual(errid, 0)
order = GetOrder(result['order_uid'])
assertIsNotNone(order)
assertFalse(order.is_dirty)
assertEqual(order.symbol, Symbol())
assertIsNotNone(order.close_time)
assertEqual(order.close_price, Ask())
assertEqual(order.tags['score'], score)
assertEqual(order.status, OrderStatus.CLOSED)

#The order was closed ?
oo = GetOpenedOrderUIDs()
for t in oo:
    assertNotEqual(t, result['order_uid'])

#close multiple orders
orders = []
for i in range(5):
    errid, result = Buy(volume=0.01, price=Ask())
    assertEqual(errid, 0)
    assertNotEqual(result['order_uid'], None)
    orders.append(result['order_uid'])
    errid = exec_command()
    assertEqual(errid, 0)
for order_uid in orders:
    order = GetOrder(order_uid)
    assertIsNotNone(order)
    assertFalse(order.is_dirty)
    assertEqual(order.status, OrderStatus.OPENED)

errid, result = CloseMultiOrders(orders)
assertEqual(errid, 0)
for order_uid in orders:
    order = GetOrder(order_uid)
    assertIsNotNone(order)
    assertEqual(order.is_dirty, RunMode() == RunModeValue.LIVE)
errid = exec_command()
assertEqual(errid, 0)
for order_uid in orders:
    order = GetOrder(order_uid)
    assertIsNotNone(order)
    assertFalse(order.is_dirty)
    assertIsNotNone(order.close_time)
    assertEqual(order.status, OrderStatus.CLOSED)

#
set_test_result("OK")
StopTester()


price=Close()
#print(f"t={TimeCurrent()}")
symbol_data = GetSymbolData(Symbol(), DefaultTimeFrame(), 35)
#print(f"wpr={iWPR(symbol_data, 5, shift=0)}")
shift = 0
c_time = symbol_data.time[shift]
if c_time is not None:
  value=iMA(symbol_data.close, timeperiod=5, matype=0, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'ma'), f'[ma]@ {c_time}')
  value=iAD(symbol_data, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'ad'), f'[ad]@ {c_time}')
  value=iADX(symbol_data, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'adx'), f'[adx]@ {c_time}')
  value=iATR(symbol_data, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'atr'), f'[atr]@ {c_time}')
  value=iBands(symbol_data.close, timeperiod=5, nbdevup=2, nbdevdn=2, matype=0, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'bands'), f'[bands]@ {c_time}')
  value=iCCI(symbol_data, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'cci'), f'[cci]@ {c_time}')
  value=iChaikin(symbol_data, fastperiod=5, slowperiod=10, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'chaikin'), f'[chaikin]@ {c_time}')
  value=iDEMA(symbol_data.close, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'dema'), f'[dema]@ {c_time}')
  value=iMomentum(symbol_data.close, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'momentum'), f'[momentum]@ {c_time}')
  value=iMFI(symbol_data, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'mfi'), f'[mfi]@ {c_time}')
  value=iMACD(symbol_data.close, fastperiod=5, slowperiod=10, signalperiod=7, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'macd'), f'[macd]@ {c_time}')
  value=iOBV(symbol_data.close, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'obv'), f'[obv]@ {c_time}')
  value=iSAR(symbol_data, acceleration=0.02, maximum=0.2, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'sar'), f'[sar]@ {c_time}')
  value=iRSI(symbol_data.close, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'rsi'), f'[rsi]@ {c_time}')
  value=iStdDev(symbol_data.close, timeperiod=5, nbdev=2, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'stddev'), f'[stddev]@ {c_time}')
  value=iStochastic(symbol_data, fastk_period=5, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'stochastic'), f'[stochastic]@ {c_time}')
  value=iTEMA(symbol_data.close, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'tema'), f'[tema]@ {c_time}')
  value=iWPR(symbol_data, timeperiod=5, shift=0)
  assertEqual(value, get_value_by_time(valid_values, c_time, 'wpr'), f'[wpr]@ {c_time}')

set_test_result("OK")